class Agent {

  Vector3D loc;
  Vector3D vel;
  Vector3D acc; 
  float maxforce;    // Maximum steering force
  float maxspeed = speed;    // Maximum speed
  
  float neighbordist;
  ArrayList neighbors;
  
  Subject subject;
  
  // for visualization
  boolean neighborhoodrepresentation;
  boolean connectivity;
  
  Agent(Vector3D l, float ms, float mf, Subject s) {
    acc = new Vector3D(0,0,0);
    vel = new Vector3D(random(-1,1),random(-1,1),random(-1,1));
    loc = l.copy();
    maxspeed = ms;
    maxforce = mf;
    subject = s;
    
    neighbors = new ArrayList();
    neighbordist = searchDistance;
    
    neighborhoodrepresentation = false;
    connectivity = false;
  }
  
  void run(ArrayList agents) 
  {
    chaseDeweyAttractor(matrix.deweyAttractor);
    updateNeighbors(agents);
    flock(neighbors);
    
    if (pause == false) 
    {
      updateLocation();
    }

    borders();
    render();
  }

  // accumulate a new acceleration each time based on three rules
  void flock(ArrayList agents) 
  {
    Vector3D sep = separate(agents);   // Separation
    Vector3D ali = align(agents);      // Alignment
    Vector3D coh = cohesion(agents);   // Cohesion
    Vector3D rand = new Vector3D(random(-1,1),random(-1,1),random(-1,1)); // Random
    
    // weight these forces
    sep.mult(separationForce);
    ali.mult(alignmentForce);
    coh.mult(cohesionForce);
    rand.mult(randomForce);
    
    // Add the force vectors to acceleration
    acc.add(sep);
    acc.add(ali);
    acc.add(coh);
    acc.add(rand);  
  }
  
  // update location
  void updateLocation() {
    // Update velocity
    vel.add(acc);
    // Limit speed
    vel.limit(maxspeed);
    loc.add(vel);
    // Reset accelertion to 0 each cycle
    acc.setXYZ(0,0,0);
  }

  void updateNeighbors(ArrayList agents) {
    // forget former neighbors
    neighbors.clear();
    
    for (int i = 0 ; i < agents.size(); i++) 
    {   
      Agent other = (Agent) agents.get(i);
      Vector3D d = new Vector3D();
      d = d.sub(loc, other.loc);
      // If the distance is greater than 0 and less than an arbitrary amount (0 when you are yourself)
      if ((d.length() > 0) && (d.length() < neighbordist))
      {
        neighbors.add(other);   
        
        
        //draw 
        if (drawNeighborConnection == true) 
        {
          noFill();
          strokeWeight(lineweight/80);
          stroke(120);
          line(loc.x,loc.y,loc.z,other.loc.x,other.loc.y,other.loc.z);   
          strokeWeight(0.2);
        }
      }
    }
  }

  void chaseDeweyAttractor(ArrayList<DeweyAttractor> DeweyAttractor) {
    for (int i = 0 ; i < DeweyAttractor.size(); i++) 
    {   
      DeweyAttractor attractor = (DeweyAttractor) DeweyAttractor.get(i);
      Vector3D d = new Vector3D();
      d = d.sub(attractor.loc, loc);
      
      if (subject.A_dewey.contains(attractor.deweyDict.id))
      {    
        // draw connection line between agent and dewey attractor
        if (drawDeweyConnection == true) 
        {
          strokeWeight(lineweight/20);
          stroke(deweyCol);
          line(loc.x,loc.y,loc.z,attractor.loc.x,attractor.loc.y,attractor.loc.z); 
        }
        
        if (gridSize/2 < d.length() && d.length() < neighbordist*3)
        {
          float weight = deweyAttractForce / d.length();
          Vector3D c = new Vector3D(d.x*weight,d.y*weight,d.z*weight);
          acc.add(c);
        }

      }
    }
  }

  void seek(Vector3D target) {
    acc.add(steer(target,false));
  }
 
  void arrive(Vector3D target) {
    acc.add(steer(target,true));
  }

  // A method that calculates a steering vector towards a target
  // Takes a second argument, if true, it slows down as it approaches the target
  Vector3D steer(Vector3D target, boolean slowdown) 
  {
    Vector3D steer;  // The steering vector
    Vector3D desired = target.sub(target,loc);  // A vector pointing from the location to the target
    float d = desired.magnitude(); // Distance from the target is the magnitude of the vector
    // If the distance is greater than 0, calc steering (otherwise return zero vector)
    if (d > 0) 
    {
      // Normalize desired
      desired.normalize();
      // Two options for desired vector magnitude (1 -- based on distance, 2 -- maxspeed)
      if ((slowdown) && (d < 100.0f)) desired.mult(maxspeed*(d/100.0f)); // This damping is somewhat arbitrary
      else desired.mult(maxspeed);
      // Steering = Desired minus Velocity
      steer = target.sub(desired,vel);
      steer.limit(maxforce);  // Limit to maximum steering force
    } 
    else 
    {
      steer = new Vector3D(0,0);
    }
    return steer;
  }
  
  void render() 
  {
    if (neighborhoodrepresentation == true)
    {
      noFill();
      stroke(150);
    }
    pushMatrix();
    strokeWeight(lineweight);
    point(loc.x,loc.y,loc.z);
    popMatrix();
  }
  
  // Wraparound
  void borders() {
    if (loc.x < 0) loc.x = worldSize.x;
    if (loc.y < 0) loc.y = worldSize.y;
    if (loc.z < 0) loc.z = worldSize.z;
    if (loc.x > worldSize.x) loc.x = 0;
    if (loc.y > worldSize.y) loc.y = 0;
    if (loc.z > worldSize.z) loc.z = 0;
 
     pushMatrix();
     noFill();
     translate(worldSize.x/2.0,worldSize.y/2.0,worldSize.z/2.0);

     strokeWeight(0.5);
     stroke(0,0,0);
     box(worldSize.x,worldSize.y,worldSize.z);
     popMatrix();
  }


  // Separation
  // if nearby agents has no co-occurrent relation, steer away
  Vector3D separate (ArrayList neighbors) {
    Vector3D sum = new Vector3D(0,0,0);
    int count = 0;
    // For every agent in the system, check if it's too close
    for (int i = 0 ; i < neighbors.size(); i++) {
      Agent other = (Agent) neighbors.get(i);
      if (!subject.A_subject.contains(other.subject.id))
      {
        float d = loc.distance(loc,other.loc);
  
        // too close? then separate!
        if (d < neighbordist) {
          // Calculate vector pointing away from neighbor
          Vector3D diff = loc.sub(loc,other.loc);
          diff.normalize();
          diff.div(d);        // Weight by distance
          sum.add(diff);
          count++;            // Keep track of how many
        }
      }
    }
    // Average -- divide by how many
    if (count > 0) {
      sum.div((float)count);
    }
    return sum;
  }
  
  
  // Alignment
  // align to nearby agents
  Vector3D align (ArrayList neighbors) {
    Vector3D sum = new Vector3D(0,0,0);
    // no neighbors, let's stop this.
    if (neighbors.size() == 0) return sum;
    
    for (int i = 0 ; i < neighbors.size(); i++) {
      Agent other = (Agent) neighbors.get(i);
      sum.add(other.vel);
      }
    sum.div((float)neighbors.size());
    sum.limit(maxforce);
    return sum;
  }

  // Cohesion
  // For the average location (i.e. center) of all nearby boids, calculate steering vector towards that location
  Vector3D cohesion (ArrayList neighbors) {
    Vector3D sum = new Vector3D(0,0,0);   // Start with empty vector to accumulate all locations
    // no neighbors, let's stop this.
    if (neighbors.size() == 0) return sum;
    for (int i = 0 ; i < neighbors.size(); i++) {
      Agent other = (Agent) neighbors.get(i);
      float d = loc.distance(loc,other.loc);
      sum.add(other.loc); // Add location
    }
    sum.div((float)neighbors.size());
    return steer(sum,false);  // Steer towards the location
  }
}
