/**
 * 
 * Version PixelView: the field of perception is scanned by the pixel in order to identify any obstacles
 *
 * Also new: a second drawing layer is introduced in order to maintain pixel-drawings
 *
 * Boid flocking implementation featuring: <br>
 * - obstacle avoidance <br>
 * - conic perception field <br>
 * - flight perturbation <br>
 * - field of perception visualization (mousecklick or 'n' for neighborhood) <br>
 * - interaction network visualization ('w' for web) <br>
 * - press 'space' to pause the simulation <br>

 * An implementation of Craig Reynold's Boids program to simulate
 * the flocking behavior of birds. Each boid steers itself based on 
 * rules of avoidance, alignment, and coherence.
 */
 
 
import processing.opengl.*;
import processing.dxf.*;
import peasy.*;
import controlP5.*;



PeasyCam cam;
ControlP5 cp5;

//global
boolean neighborhoodview;

//private
int mX,mY,camX,camY;

//world size
Vector3D worldSize = new Vector3D(gridSize*11,gridSize*11,gridSize*11);

Flock flock;
Matrix matrix;
Random rand = new Random();

void setup() {
  
  ReadData();
  
  size(1000,1000,P3D);
  colorMode(RGB,255,255,255,100);
  
  cam = new PeasyCam(this, 200);
  cam.setMinimumDistance(-1200);
  cam.setMaximumDistance(1200);

  // Set up the GUI
  cp5 = new ControlP5(this);
  setGui();

  noStroke();
  
  flock = new Flock();
  // Add an initial set of boids into the system
  for (int i = 0; i < boidNumber; i++) 
  {
    flock.addAgent(new Agent(new Vector3D(random(width/2),random(height/2)),2.0f,0.05f,subjects.get(rand.nextInt(subjects.size()))));
  }
  
  matrix = new Matrix();
  matrix.setupDeweyAttractor();
}



void draw() 
{
  background(255);
    
  matrix.render();
  flock.run();

  //GUI
  gui();
}

public void pause()
{
  pause = !pause;
}
